----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona, Norwegian University of Science and Technology (NTNU)
-- Engineer-1: Assistant Professor at ECE-department, College of Engineering 
-- Engineer-2: Professor at Department of Information Security and Communication Technology
-- 
-- Create Date: 06/22/2019 02:06:31 PM
-- Design Name-1: Mohamed El-Hadedy, also Airforce Research Lab(AFRL) fellow, US-AFA
-- Design Name-2: Danilo Gligoroski, also Visiting professor: Department of Mathematics and Statistics, University of South Florida, USA

-- Module Name: Shuffler - Behavioral
-- Project Name: Gage Hash Function , C: 224 and r: 8
-- Target Devices: xc7a35tcpg236-1 
-- Tool Versions: Vivado 2019.1
-- Description: Hash function Engine (FIFO for storing the message coming from the main testbench (can be coming from any serial communication protocol as PCI-E or UART, USB, SPI, I2C or more)
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity BIT_SHUFFLER is
    generic (c:positive:=224 ; r:positive:= 8);

    Port ( clk    : in  std_logic; 
           rst    : in  std_logic;
           OS_POS : in  STD_LOGIC_VECTOR (integer(ceil(log2(real(c+r))))-1 downto 0);
           NS_POS : out STD_LOGIC_VECTOR (integer(ceil(log2(real(c+r))))-1 downto 0)
         );
end BIT_SHUFFLER;

architecture Behavioral of BIT_SHUFFLER is

type mem_shuffle is array(0 to (2**(integer(ceil(log2(real(c+r))))))-1) of std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0); 

signal shf_mem: mem_shuffle:=(
                               std_logic_vector(to_unsigned(176,8)), 
                               std_logic_vector(to_unsigned(190,8)), 
                               std_logic_vector(to_unsigned(193,8)),
                               std_logic_vector(to_unsigned(207,8)),
                               std_logic_vector(to_unsigned(210,8)),
                               std_logic_vector(to_unsigned(221,8)),
                               std_logic_vector(to_unsigned(228,8)),
                               std_logic_vector(to_unsigned(3,8)),
                               std_logic_vector(to_unsigned(184,8)),
                               std_logic_vector(to_unsigned(198,8)),
                               std_logic_vector(to_unsigned( 201,8)),
                               std_logic_vector(to_unsigned( 215,8)),
                               std_logic_vector(to_unsigned( 218,8)),
                               std_logic_vector(to_unsigned( 229,8)),
                               std_logic_vector(to_unsigned( 4,8)),
                               std_logic_vector(to_unsigned( 11,8)),
                               std_logic_vector(to_unsigned( 192,8)),
                               std_logic_vector(to_unsigned( 206,8)),
                               std_logic_vector(to_unsigned( 209,8)),
                               std_logic_vector(to_unsigned( 223,8)),
                               std_logic_vector(to_unsigned( 226,8)),
                               std_logic_vector(to_unsigned( 5,8)),
                               std_logic_vector(to_unsigned( 12,8)),
                               std_logic_vector(to_unsigned( 19,8)),
                               std_logic_vector(to_unsigned( 200,8)),
                               std_logic_vector(to_unsigned( 214,8)),
                               std_logic_vector(to_unsigned( 217,8)),
                               std_logic_vector(to_unsigned( 231,8)),
                               std_logic_vector(to_unsigned( 2,8)),
                               std_logic_vector(to_unsigned( 13,8)),
                               std_logic_vector(to_unsigned( 20,8)),
                               std_logic_vector(to_unsigned( 27,8)),
                               std_logic_vector(to_unsigned( 208,8)),
                               std_logic_vector(to_unsigned( 222,8)),
                               std_logic_vector(to_unsigned( 225,8)),
                               std_logic_vector(to_unsigned( 7,8)),
                               std_logic_vector(to_unsigned( 10,8)),
                               std_logic_vector(to_unsigned( 21,8)),
                               std_logic_vector(to_unsigned( 28,8)),
                               std_logic_vector(to_unsigned( 35,8)),
                               std_logic_vector(to_unsigned( 216,8)),
                               std_logic_vector(to_unsigned( 230,8)),
                               std_logic_vector(to_unsigned( 1,8)),
                               std_logic_vector(to_unsigned( 15,8)),
                               std_logic_vector(to_unsigned( 18,8)),
                               std_logic_vector(to_unsigned( 29,8)),
                               std_logic_vector(to_unsigned( 36,8)),
                               std_logic_vector(to_unsigned( 43,8)),
                               std_logic_vector(to_unsigned( 224,8)),
                               std_logic_vector(to_unsigned( 6,8)),
                               std_logic_vector(to_unsigned( 9,8)),
                               std_logic_vector(to_unsigned( 23,8)),
                               std_logic_vector(to_unsigned( 26,8)),
                               std_logic_vector(to_unsigned( 37,8)),
                               std_logic_vector(to_unsigned( 44,8)),
                               std_logic_vector(to_unsigned( 51,8)),
                               std_logic_vector(to_unsigned( 0,8)),
                               std_logic_vector(to_unsigned( 14,8)),
                               std_logic_vector(to_unsigned( 17,8)),
                               std_logic_vector(to_unsigned( 31,8)),
                               std_logic_vector(to_unsigned( 34,8)),
                               std_logic_vector(to_unsigned( 45,8)),
                               std_logic_vector(to_unsigned( 52,8)),
                               std_logic_vector(to_unsigned( 59,8)),
                               std_logic_vector(to_unsigned( 8,8)),
                               std_logic_vector(to_unsigned( 22,8)),
                               std_logic_vector(to_unsigned( 25,8)),
                               std_logic_vector(to_unsigned( 39,8)),
                               std_logic_vector(to_unsigned( 42,8)),
                               std_logic_vector(to_unsigned( 53,8)),
                               std_logic_vector(to_unsigned( 60,8)),
                               std_logic_vector(to_unsigned( 67,8)),
                               std_logic_vector(to_unsigned( 16,8)),
                               std_logic_vector(to_unsigned( 30,8)),
                               std_logic_vector(to_unsigned( 33,8)),
                               std_logic_vector(to_unsigned( 47,8)),
                               std_logic_vector(to_unsigned( 50,8)),
                               std_logic_vector(to_unsigned( 61,8)),
                               std_logic_vector(to_unsigned( 68,8)),
                               std_logic_vector(to_unsigned( 75,8)),
                               std_logic_vector(to_unsigned( 24,8)),
                               std_logic_vector(to_unsigned( 38,8)),
                               std_logic_vector(to_unsigned( 41,8)),
                               std_logic_vector(to_unsigned( 55,8)),
                               std_logic_vector(to_unsigned( 58,8)),
                               std_logic_vector(to_unsigned( 69,8)),
                               std_logic_vector(to_unsigned( 76,8)),
                               std_logic_vector(to_unsigned( 83,8)),
                               std_logic_vector(to_unsigned( 32,8)),
                               std_logic_vector(to_unsigned( 46,8)),
                               std_logic_vector(to_unsigned( 49,8)),
                               std_logic_vector(to_unsigned( 63,8)),
                               std_logic_vector(to_unsigned( 66,8)),
                               std_logic_vector(to_unsigned( 77,8)),
                               std_logic_vector(to_unsigned( 84,8)),
                               std_logic_vector(to_unsigned( 91,8)),
                               std_logic_vector(to_unsigned( 40,8)),
                               std_logic_vector(to_unsigned( 54,8)),
                               std_logic_vector(to_unsigned( 57,8)),
                               std_logic_vector(to_unsigned( 71,8)),
                               std_logic_vector(to_unsigned( 74,8)),
                               std_logic_vector(to_unsigned( 85,8)),
                               std_logic_vector(to_unsigned( 92,8)),
                               std_logic_vector(to_unsigned( 99,8)),
                               std_logic_vector(to_unsigned( 48,8)),
                               std_logic_vector(to_unsigned( 62,8)),
                               std_logic_vector(to_unsigned( 65,8)),
                               std_logic_vector(to_unsigned( 79,8)),
                               std_logic_vector(to_unsigned( 82,8)),
                               std_logic_vector(to_unsigned( 93,8)),
                               std_logic_vector(to_unsigned( 100,8)),
                               std_logic_vector(to_unsigned( 107,8)),
                               std_logic_vector(to_unsigned( 56,8)),
                               std_logic_vector(to_unsigned( 70,8)),
                               std_logic_vector(to_unsigned( 73,8)),
                               std_logic_vector(to_unsigned( 87,8)),
                               std_logic_vector(to_unsigned( 90,8)),
                               std_logic_vector(to_unsigned( 101,8)),
                               std_logic_vector(to_unsigned( 108,8)),
                               std_logic_vector(to_unsigned( 115,8)),
                               std_logic_vector(to_unsigned( 64,8)),
                               std_logic_vector(to_unsigned( 78,8)),
                               std_logic_vector(to_unsigned( 81,8)),
                               std_logic_vector(to_unsigned( 95,8)),
                               std_logic_vector(to_unsigned( 98,8)),
                               std_logic_vector(to_unsigned( 109,8)),
                               std_logic_vector(to_unsigned( 116,8)),
                               std_logic_vector(to_unsigned( 123,8)),
                               std_logic_vector(to_unsigned( 72,8)),
                               std_logic_vector(to_unsigned( 86,8)),
                               std_logic_vector(to_unsigned(89,8)),
                               std_logic_vector(to_unsigned( 103,8)),
                               std_logic_vector(to_unsigned( 106,8)),
                               std_logic_vector(to_unsigned( 117,8)),
                               std_logic_vector(to_unsigned( 124,8)),
                               std_logic_vector(to_unsigned( 131,8)),
                               std_logic_vector(to_unsigned( 80,8)),
                               std_logic_vector(to_unsigned( 94,8)),
                               std_logic_vector(to_unsigned( 97,8)),
                               std_logic_vector(to_unsigned( 111,8)),
                               std_logic_vector(to_unsigned( 114,8)),
                               std_logic_vector(to_unsigned( 125,8)),
                               std_logic_vector(to_unsigned( 132,8)),
                               std_logic_vector(to_unsigned( 139,8)),
                               std_logic_vector(to_unsigned( 88,8)),
                               std_logic_vector(to_unsigned( 102,8)),
                               std_logic_vector(to_unsigned( 105,8)),
                               std_logic_vector(to_unsigned( 119,8)),
                               std_logic_vector(to_unsigned( 122,8)),
                               std_logic_vector(to_unsigned( 133,8)),
                               std_logic_vector(to_unsigned( 140,8)),
                               std_logic_vector(to_unsigned( 147,8)),
                               std_logic_vector(to_unsigned( 96,8)),
                               std_logic_vector(to_unsigned( 110,8)),
                               std_logic_vector(to_unsigned( 113,8)),
                               std_logic_vector(to_unsigned( 127,8)),
                               std_logic_vector(to_unsigned( 130,8)),
                               std_logic_vector(to_unsigned( 141,8)),
                               std_logic_vector(to_unsigned( 148,8)),
                               std_logic_vector(to_unsigned( 155,8)),
                               std_logic_vector(to_unsigned( 104,8)),
                               std_logic_vector(to_unsigned( 118,8)),
                               std_logic_vector(to_unsigned( 121,8)),
                               std_logic_vector(to_unsigned( 135,8)),
                               std_logic_vector(to_unsigned( 138,8)),
                               std_logic_vector(to_unsigned( 149,8)),
                               std_logic_vector(to_unsigned( 156,8)),
                               std_logic_vector(to_unsigned( 163,8)),
                               std_logic_vector(to_unsigned( 112,8)),
                               std_logic_vector(to_unsigned( 126,8)),
                               std_logic_vector(to_unsigned( 129,8)),
                               std_logic_vector(to_unsigned( 143,8)),
                               std_logic_vector(to_unsigned( 146,8)),
                               std_logic_vector(to_unsigned( 157,8)),
                               std_logic_vector(to_unsigned( 164,8)),
                               std_logic_vector(to_unsigned( 171,8)),
                               std_logic_vector(to_unsigned( 120,8)),
                               std_logic_vector(to_unsigned( 134,8)),
                               std_logic_vector(to_unsigned( 137,8)),
                               std_logic_vector(to_unsigned( 151,8)),
                               std_logic_vector(to_unsigned( 154,8)),
                               std_logic_vector(to_unsigned( 165,8)),
                               std_logic_vector(to_unsigned( 172,8)),
                               std_logic_vector(to_unsigned( 179,8)),
                               std_logic_vector(to_unsigned( 128,8)),
                               std_logic_vector(to_unsigned( 142,8)),
                               std_logic_vector(to_unsigned( 145,8)),
                               std_logic_vector(to_unsigned( 159,8)),
                               std_logic_vector(to_unsigned( 162,8)),
                               std_logic_vector(to_unsigned( 173,8)),
                               std_logic_vector(to_unsigned( 180,8)),
                               std_logic_vector(to_unsigned( 187,8)),
                               std_logic_vector(to_unsigned( 136,8)),
                               std_logic_vector(to_unsigned( 150,8)),
                               std_logic_vector(to_unsigned( 153,8)),
                               std_logic_vector(to_unsigned( 167,8)),
                               std_logic_vector(to_unsigned( 170,8)),
                               std_logic_vector(to_unsigned( 181,8)),
                               std_logic_vector(to_unsigned( 188,8)),
                               std_logic_vector(to_unsigned( 195,8)),
                               std_logic_vector(to_unsigned( 144,8)),
                               std_logic_vector(to_unsigned( 158,8)),
                               std_logic_vector(to_unsigned( 161,8)),
                               std_logic_vector(to_unsigned( 175,8)),
                               std_logic_vector(to_unsigned( 178,8)),
                               std_logic_vector(to_unsigned( 189,8)),
                               std_logic_vector(to_unsigned( 196,8)),
                               std_logic_vector(to_unsigned( 203,8)),
                               std_logic_vector(to_unsigned( 152,8)),
                               std_logic_vector(to_unsigned( 166,8)),
                               std_logic_vector(to_unsigned( 169,8)),
                               std_logic_vector(to_unsigned( 183,8)),
                               std_logic_vector(to_unsigned( 186,8)),
                               std_logic_vector(to_unsigned( 197,8)),
                               std_logic_vector(to_unsigned( 204,8)),
                               std_logic_vector(to_unsigned( 211,8)),
                               std_logic_vector(to_unsigned( 160,8)),
                               std_logic_vector(to_unsigned( 174,8)),
                               std_logic_vector(to_unsigned( 177,8)),
                               std_logic_vector(to_unsigned( 191,8)),
                               std_logic_vector(to_unsigned( 194,8)),
                               std_logic_vector(to_unsigned( 205,8)),
                               std_logic_vector(to_unsigned( 212,8)),
                               std_logic_vector(to_unsigned( 219,8)),
                               std_logic_vector(to_unsigned( 168,8)),
                               std_logic_vector(to_unsigned( 182,8)),
                               std_logic_vector(to_unsigned( 185,8)),
                               std_logic_vector(to_unsigned( 199,8)),
                               std_logic_vector(to_unsigned( 202,8)),
                               std_logic_vector(to_unsigned( 213,8)),
                               std_logic_vector(to_unsigned( 220,8)),
                               std_logic_vector(to_unsigned( 227,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)),
                               std_logic_vector(to_unsigned(0,8)) 
                              );                             
begin

  --  process(clk,rst) 
  --  begin 
  --      if(rst = '1')then 
  --       NS_POS<= (others =>'0');
  --      elsif(rising_edge(clk)) then 
         NS_POS<=shf_mem(conv_integer(OS_POS));

   --     end if;  
   -- end process; 
end Behavioral;
