----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona 
-- Engineer: Assistant Professor, ECE-department 
-- 
-- Create Date: 07/10/2019 04:12:37 PM
-- Design Name: Mohamed El-Hadedy (Aly)
-- Module Name: Engine_Q - Behavioral
-- Project Name: Gage Hash Function 
-- Target Devices: Basys-3
-- Tool Versions: Vivado 2019.1
-- Description: Smart Engine
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity Engine_Q is
    Port ( clk       : in  STD_LOGIC;
           rst       : in  STD_LOGIC;
           STRT      : in  STD_LOGIC; 
           L         : in  STD_LOGIC_VECTOR(1 downto 0);
           OSTATE    : in  STD_LOGIC;
           INPQ      : out STD_LOGIC;
           QEN_OUT   : out STD_LOGIC;
           RAM_WR    : out STD_LOGIC
         );
end Engine_Q;

architecture Behavioral of Engine_Q is

signal D1, D2, D3, D4: std_logic; 
signal INPQ_SEL    : std_logic_vector(3 downto 0);

component Q_sbox 
  Port   (
         clk    : in  std_logic;
         rst    : in  std_logic; 
         EN_IN  : in  std_logic; 
         INP    : in  std_logic_vector(3 downto 0);
         OP     : out std_logic;
         EN_OUT : out std_logic;
         RAM_OUT: out std_logic    
         );
end component;

signal INPQ_TMP     : std_logic_vector(3 downto 0); 
signal X0, X1,X2, X3, x4, x5: std_logic; 
signal STATE_SEL    : std_logic;
signal v0, v1, v2,v3: std_logic; 
signal LSEL         : std_logic; 
signal tmp          : std_logic;
signal tmp0,tmp1    : std_logic; 
signal QEN_OUT_FSM  : std_logic;

type Q_fsm_states is (
    INIT,
    LEADER,
    STATE_L1,
    STATE_L2
  );
  -- signals declaration
  signal q_reg_fsm, q_nxt_fsm   : Q_fsm_states;

begin


-- REPLACING MUX_PROC by two combinational state

DELAY_0: process(clk,rst)
         begin 
            if (rst = '1') then 
                v0 <= '0';
            elsif (rising_edge(clk)) then 
                if (X0 = '1') then 
                  v0 <= OSTATE;
                end if;  
            end if;
         end process; 
DELAY_1: process(clk,rst)
         begin 
            if (rst = '1') then 
                v1 <= '0';
            elsif (rising_edge(clk)) then 
                if (X1 = '1') then 
                  v1 <= OSTATE;
                end if;  
            end if;
         end process;         
         
tmp0 <= ((STATE_SEL and v1 ) or (not STATE_SEL and v0)); 
         
D1 <= ((LSEL and tmp0) or (not LSEL and L(1))); 
         
         
         
DELAY_2: process(clk,rst)
         begin 
            if (rst = '1') then 
                v2 <= '0';
            elsif (rising_edge(clk)) then 
                if (X2 = '1') then 
                  v2 <= OSTATE;
                end if;  
            end if;
         end process;  
DELAY_3: process(clk,rst)
         begin 
            if (rst = '1') then 
                v3 <= '0';
            elsif (rising_edge(clk)) then 
                if (X3 = '1') then 
                  v3 <= OSTATE;
                end if;  
            end if;
         end process;          
                 
tmp1 <= ((STATE_SEL and v3) or (not STATE_SEL and v2)); 
D2 <= ((LSEL and tmp1) or (not LSEL and L(0))); 


DELAY_D4: process(clk,rst)
         begin 
            if (rst = '1') then 
                   D3 <= '0';
            elsif (rising_edge(clk)) then 
                if (X4 = '1') then 
                   D3 <= OSTATE;  
                end if; 
            end if;
         end process;       

DELAY_D5: process(clk,rst)
         begin 
            if (rst = '1') then 
                   D4 <= '0';
            elsif (rising_edge(clk)) then
                if (X5 = '1') then  
                   D4 <= OSTATE;  
                end if; 
            end if;
         end process; 




                 INPQ_TMP(3) <= D1;
                 INPQ_TMP(2) <= D2;
                 INPQ_TMP(1) <= D3;
                 INPQ_TMP(0) <= D4;

                 Q_GEN: Q_sbox port map (
                   clk     => clk,
                   rst     => rst,
                   EN_IN   => QEN_OUT_FSM,
                   INP     => INPQ_TMP, 
                   op      => INPQ, 
                   EN_OUT  => tmp,
                   RAM_OUT => RAM_WR
                   );
QEN_OUT <= tmp;

FSM: process(clk, rst)
     begin 
    if (rst = '1') then
        LSEL <= '0';
        X0   <= '0';
        X1   <= '0';
        X2   <= '0';
        X3   <= '0';
        X4   <= '0';
        X5   <= '0';
        STATE_SEL <= '0'; 
        QEN_OUT_FSM   <= '0';
        INPQ_SEL  <= (others =>'0');
        
     elsif(rising_edge(clk)) then
 
      if (STRT = '1') then 
        LSEL      <= '0';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '0';
        X5        <= '0';
       elsif  (INPQ_SEL = 0) then
        LSEL      <= '0';
        STATE_SEL <= '0';
        X0        <= '1';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '1';
        X5        <= '0'; 
       INPQ_SEL <= INPQ_SEL + 1;
     elsif  (INPQ_SEL = 1) then
        LSEL      <= '0';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '0';
        X2        <= '1';
        X3        <= '0';
        X4        <= '0';
        X5        <= '1'; 
       INPQ_SEL <= INPQ_SEL + 1;
     elsif  (INPQ_SEL = 2) then
        LSEL      <= '0';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '0';
        X5        <= '0'; 
        QEN_OUT_FSM   <= '1';
       INPQ_SEL <= INPQ_SEL + 1;               
     elsif  (INPQ_SEL = 3) then
        LSEL      <= '1';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '1';
        X2        <= '0';
        X3        <= '0';
        X4        <= '1';
        X5        <= '0'; 
        QEN_OUT_FSM   <= '0';

       INPQ_SEL <= INPQ_SEL + 1;             
     elsif  (INPQ_SEL = 4) then
        LSEL      <= '1';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '0';
        X2        <= '0';
        X3        <= '1';
        X4        <= '0';
        X5        <= '1'; 
        QEN_OUT_FSM   <= '0';
       INPQ_SEL <= INPQ_SEL + 1;                      
     elsif  (INPQ_SEL = 5) then
        LSEL      <= '1';
        STATE_SEL <= '0';
        X0        <= '0';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '0';
        X5        <= '0'; 
        QEN_OUT_FSM   <= '1';
       INPQ_SEL <= INPQ_SEL + 1;               
     elsif  (INPQ_SEL = 6) then
        LSEL      <= '1';
        STATE_SEL <= '1';
        X0        <= '1';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '1';
        X5        <= '0'; 
        QEN_OUT_FSM   <= '0';
       INPQ_SEL <= INPQ_SEL + 1;               
     elsif  (INPQ_SEL = 7) then
        LSEL      <= '1';
        STATE_SEL <= '1';
        X0        <= '0';
        X1        <= '0';
        X2        <= '1';
        X3        <= '0';
        X4        <= '0';
        X5        <= '1'; 
        QEN_OUT_FSM   <= '0';
       INPQ_SEL <= INPQ_SEL + 1;               
     elsif  (INPQ_SEL = 8) then
        LSEL      <= '1';
        STATE_SEL <= '1';
        X0        <= '0';
        X1        <= '0';
        X2        <= '0';
        X3        <= '0';
        X4        <= '0';
        X5        <= '0'; 
        QEN_OUT_FSM   <= '1';
        INPQ_SEL <= INPQ_SEL -5;        

      end if; 
      
     
     end if; 
     
     end process; 



end Behavioral;
